
;+

function trace_mdi_align,tindex,tdata,mindex,mdata, $
                         trace_mask=rmaskin,mdi_mask=tmaskin, $
                         tmdi2traceguess=tmdi2traceguess, $
                         xvr=xvr,yvr=yvr,xvi=xvi,yvi=yvi, $
                         quiet=quiet,mdi2trace=mdi2trace, $
                         new_mdata=inew,new_tdata=rimage, $
                         blindguess=blindguess, xytrace=xytrace, $
                         _EXTRA=extra

;NAME:
;     TRACE_MDI_ALIGN
;PURPOSE:
;     Use an MDI white light image to fix the pointing in a TRACE
;     white light image by assuming that the MDI pointing is correct.
;CATEGORY:
;CALLING SEQUENCE:
;     new_trace_index = trace_mdi_align(tindex,tdata,mindex,mdata)
;INPUTS:
;     tindex  = TRACE index (a structure)
;     tdata   = TRACE image
;     mindex  = MDI index (a structure)
;     mdata   = MDI image
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS:
;     trace_mask = Index giving points in the TRACE image to neglect in the
;                  alignment.
;     mdi_mask = Index giving points in the MDI image to neglect in the
;                alignment.
;     mdi2trace = the final transformation from the MDI image to the TRACE
;                 image (output).  You probably don't really want this.
;     new_tdata = the piece of the TRACE image used to do the alignment
;                 (output).
;     new_mdata = the piece of the MDI images transformed onto new_tdata
;                 (output).
;     /quiet = work quietly
;     _extra = Any extra kewords are passed to auto_align_image.  E.g. scale,
;             /amoeba, etc.
;
;     INPUT PARAMETERS FOR THE INITIAL GUESS (NOT REQUIRED):
;
;     These keywords allow you to avoid clicking on points in the two images
;     which may be useful for automated alignments etc. 
;
;     /blindguess = If set, use the alignment in the MDI and TRACE headers
;                   to set the initial guess: do not ask the user to click
;                   on points in the two images.  If this works, fine, but if
;                   the guess is too far off, it is possible that
;                   auto_align_images will fail.  If /blindguess is set, the
;                   tmdi2trace and xvr, xvi etc. keywords are ignored (but
;                   will be set on output).
;     xytrace = 4 element array giving the LL and UR corners of a subregion
;               to use in the alignment.  If this is not set, you will be
;               asked to define the subregion with the mouse. [x0,x1,y0,y1]
;               Ignored if /blindguess is not set.
;
;     If the following are passed as input, they must all be passed in and they must
;     all be consistent.  I.e., the transformation must apply to the subimages
;     defined by xvr,yvr,xvi,yvi.  If they are not consistent, the answer you 
;     get will be incorrect.  Garbage in, garbage out.  Be careful with these,
;     getting them right can be a bit tricky!  If these are not set on input,
;     they will be set on output so that they can be used for a subsequent call.
;
;     tmdi2traceguess = Tranform vector (from caltrans.pro) giving the initial
;                       guess for the alignment.  If this is not specified,
;                       the user will be prompted to click on corresponding
;                       points in the TRACE and MDI images in order to get
;                       the initial guess.  If this is passed, xvr,yvr,xvi,yvi
;                       must also be passed in.  If undefined, it will be
;                       set on output to the guess from setpts_roi.
;     xvr,yvr = 2 element arrays giving LL a UR corners of the TRACE subimage
;               to use.  Ignored unless tmdi2traceguess is also passed.
;               If undefined, it will be set on output to the guess from 
;               setpts_roi.
;     xvi,yvi = 2 element arrays giving LL a UR corners of the MDI subimage
;               to use.  Ignored unless tmdi2traceguess is also passed.
;               If undefined, it will be set on output to the guess from 
;               setpts_roi.
;OUTPUTS:
;     new_trace_index = trace index with the pointing (i.e. .xcen and .ycen)
;                       corrected.  .xcen,.ycen,.crval and .crpix  are the
;                       only things changed in the new index.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     VERY SLOW
;
;     Obviously the TRACE and MDI images should be close in time to avoid
;     evolution in the AR (but solar rotation and foreshortening are 
;     accounted for).
;
;     MDI pointing is assumed to be correct.
;PROCEDURE:
;     The user first gives an initial guess for the transformation between
;     TRACE and MDI using setpts_roi and then the guess is refined using 
;     auto_align_image.   The transformation is then used to set the pointing
;     of TRACE assuming that the MDI pointing is correct.
;EXAMPLE:
;     ; First get the MDI data
;     mdata  = readfits('Some_MDI_FITS_File',smdihead)
;     mindex = fitshead2struct(smdihead)
;     ; Second get the TRACE data
;     trace_cat,some_start_time,some_end_time,cat
;     ss = where(cat.wave_len EQ 'WL')
;     trace_cat2data,cat[ss[0]],i,d
;     trace_prep,i,d,tindex,tdata,/wave2point
;     ; Third align the MDI and TRACE data
;     tindex_new = trace_mdi_align(tindex,tdata,mindex,mdata)
;MODIFICATION HISTORY:
;     T. Metcalf 2001-Oct-17
;     T. Metcalf 2001-Nov-14 Added a warning for a large coordinate shift.
;     T. Metcalf 2001-Nov-15 Also update crpix1,crpix2,crval1,crval2 so that
;                            the new index with the new pointing will work
;                            with index2map and plot_map which gets pointing
;                            from crpix and crval rather than xcen and ycen.
;     T. Metcalf 2001-Nov-19 Added the /blindguess keyword.  Also mask
;                            off-limb points.
;     T. Metcalf 2001-Apr-04 Added correction for parallax in longitude
;                            between the earth and SOHO points of view
;                            (a small correction).  Parallax in latitude
;                            was already accounted for.
;     T. Metcalf 2001-Apr-23 Cleaned up code a bit.
;-

if strtrim(strupcase(tindex.wave_len),2) NE 'WL' then $
   message,'The TRACE image is not white light'

nxmdi = n_elements(mdata[*,0])
nymdi = n_elements(mdata[0,*])

bad = where(mdata LT mindex.d_min0 OR mdata GT mindex.d_max0,nbad)
if nbad GT 0 then mdata[bad] = 0

; mdidate should be equivalent time at earth but the time at L1 and the
; time observed at earth only differ by 5 seconds or so
; and I'm not sure if date_obs already has the 5 sec correction, so ignore it
; here since differential rotation is small over 5 seconds anyway. 

mdidate    = anytim(mindex.date_obs,/yoh)  
tracedate0 = anytim(tindex.date_obs,/yoh)

; Get the pixel size of MDI image

mdipix = mindex.im_scale
; This corrects the known problem with the MDI plate scale.  It might be
; necessary to update the value periodically???
if abs(mdipix-1.9778400) LT 0.00001 then begin
   mdipix = 1.982
   message,/info,'Correcting MDI plate scale of 1.97784 to '+string(mdipix)
endif
mdipix_earth = mdipix*mindex.EARTH_R0/mindex.OBS_R0

nxtrace = long(tindex.naxis1)
nytrace = long(tindex.naxis2)

; The first task is to get an initial guess for the MDI->TRACE
; transformation.  If tmdi2traceguess is passed in, we are all set.  If
; /blindguess is set, then use the TRACE and MDI pointing info in tindex and
; mindex.  Otherwise prompt the user to select ROI's and click on corresponding 
; points in the subimages.

if keyword_set(blindguess) then begin
   if n_elements(xytrace) NE 4 then begin
      nxr = n_elements(tdata[*,0])
      nyr = n_elements(tdata[0,*])
      binr = 1L+max([nxr,nyr])/640L
      REPEAT nxr = nxr - 1L UNTIL nxr MOD binr EQ 0
      REPEAT nyr = nyr - 1L UNTIL nyr MOD binr EQ 0
      ; Select a smaller FOV which encompasses only the features to be
      ; aligned.  This is for speed; it'd be OK to take the whole FOV.
      message,/info,'Select a sub region for the alignment.  The smaller the '+$
                    'region, the faster the calculation.'
      junk = wdefroi(rebin(tdata[0:nxr-1,0:nyr-1],nxr/binr,nyr/binr,/sample), $
                     xvroi,yvroi)
      xytrace = [min(xvroi),max(xvroi),min(yvroi),max(yvroi)]*binr
   endif
   xvr = [xytrace[0]>0,xytrace[1]<(nxtrace-1L)]
   yvr = [xytrace[2]>0,xytrace[3]<(nytrace-1L)]
   ; Now get the equivalent region in the MDI data
   ;xcen,ycen are the coordinates of the center of the FOV (arcsec)
   xcen = tindex.xcen + (xvr[0]/2.0 + (xvr[1]-(nxtrace-1L))/2.0)*tindex.cdelt1
   ycen = tindex.ycen + (yvr[0]/2.0 + (yvr[1]-(nytrace-1L))/2.0)*tindex.cdelt2
   ; Correct these for solar rotation
   ll = xy2lonlat([xcen,ycen],tracedate0)
   ddays = int2secarr(mdidate,tracedate0)/86400.
   ll[0] = ll[0] + diff_rot(ddays,ll[1],/allen)
   xy = lonlat2xy(ll,mdidate)
   xcen = xy[0]
   ycen = xy[1]
   ; Now convert to MDI pixel coordinates
   xvi = fltarr(2) & yvi = fltarr(2)
   xvi[0] = mindex.crpix1+xcen/mdipix_earth - $
           ((xvr[1]-xvr[0])/2)*tindex.cdelt1/mdipix_earth
   xvi[1] = mindex.crpix1+xcen/mdipix_earth + $
           ((xvr[1]-xvr[0])/2)*tindex.cdelt1/mdipix_earth
   yvi[0] = mindex.crpix2+ycen/mdipix_earth - $
           ((yvr[1]-yvr[0])/2)*tindex.cdelt2/mdipix_earth
   yvi[1] = mindex.crpix2+ycen/mdipix_earth + $
           ((yvr[1]-yvr[0])/2)*tindex.cdelt2/mdipix_earth
   ; Shift by the fractional part of xvi,yvi
   xshift = xvi[0]-long(xvi[0]) & yshift = yvi[0]-long(yvi[0]) 
   xvi = long(xvi)
   yvi = long(yvi)
   ; Make sure the boundaries are not out of range
   if xvi[0] LT 0 then begin
      xshift = xshift - xvi[0]
      xvi[0] = 0
   endif
   if yvi[0] LT 0 then begin
      yshift = yshift - yvi[0]
      yvi[0] = 0
   endif
   if xvi[1] GT nxmdi-1 then begin
      xvi[1] = nxmdi-1
   endif
   if yvi[1] GT nymdi-1 then begin
      yvi[1] = nymdi-1
   endif
   ; Get the transformation
   tmdi2traceguess = rss2pq(xshift=xshift,yshift=yshift,rot12=0.0, $
                            xscale=mdipix_earth/tindex.cdelt1, $
                            yscale=mdipix_earth/tindex.cdelt2)
endif

if n_elements(tmdi2traceguess) LE 0 then begin

   ; Get the initial guess from the user if one
   ; is not passed in and /blindguess is not set

   message,/info,'Select the regions of the TRACE and MDI images to ' + $
                 'analyze.  Then click on corresponding points in the two ' + $
                 'images to set the initial guess for the transformation ' + $
                 'between them.'

   transtype = {type:'s', $
                ref_x:tindex.cdelt1, $
                ref_y:tindex.cdelt2, $
                pix_x:mdipix_earth, $
                pix_y:mdipix_earth, $
                transpose:0,phi:0.0}

   pp = setpts_roi(mdata,tdata,key=transtype, $
                   xvi=xvi,yvi=yvi,xvr=xvr,yvr=yvr)

   ; Get back to the restricted FOV since we'll use that in the 
   ; alignment

   pp(0,0,*) = pp(0,0,*) -  xvr(0)  ; x-coord of ref image
   pp(1,0,*) = pp(1,0,*) -  yvr(0)  ; y-coord of ref image
   pp(0,1,*) = pp(0,1,*) -  xvi(0)  ; x-coord of transformed image
   pp(1,1,*) = pp(1,1,*) -  yvi(0)  ; y-coord of transformed image

   tmdi2traceguess = caltrans(pp,transtype=transtype)
   if NOT keyword_set(quiet) then begin
      print,'Restricted initial guess for transformation:'
      pq2rss,tmdi2traceguess(*,0),tmdi2traceguess(*,1), $
         erot,exscl,eyscl,exshft,eyshft,enrss
      print
   endif
endif

rimage = tdata[xvr[0]:xvr[1],yvr[0]:yvr[1]]
timage = mdata[xvi[0]:xvi[1],yvi[0]:yvi[1]]

x0 = tindex.xcen/tindex.cdelt1 + ((xvr[1]+xvr[0])/2.0 - (tindex.naxis1-1.)/2.)
y0 = tindex.ycen/tindex.cdelt2 + ((yvr[1]+yvr[0])/2.0 - (tindex.naxis2-1.)/2.)
b0 = get_rb0p(tracedate0,/b0angle)
ll = get_latlon(nx=xvr[1]-xvr[0]+1L, $
                ny=yvr[1]-yvr[0]+1L, $
                x0=x0, $
                y0=y0, $
                b0=b0, $
                pang=0.0, $ ; Assumes TRACE up is north
                pixs=[tindex.cdelt1,tindex.cdelt2], $
                rad=get_rb0p(tracedate0,/radius))

rho = acos(cos(ll.latitude*!dtor)*cos(b0)*cos(ll.cmd*!dtor) + $
           sin(ll.latitude*!dtor)*sin(b0))
offlimb = where(rho GT 85.*!dtor,nofflimb)
delvarx,ll,rho

rmask = tdata & rmask[*] = 0.0
if n_elements(rmaskin) GT 0 then rmask[rmaskin] = 1.0
rmask = rmask[xvr[0]:xvr[1],yvr[0]:yvr[1]]
if nofflimb GT 0 then begin
   rmask[offlimb] = 1.0
   if NOT keyword_set(quiet) then $
      message,/info,strcompress('Masking out '+string(nofflimb)+' off-limb points.')
endif
rmask = where(rmask,nrmask)
if nrmask LE 0 then delvarx,rmask

if  n_elements(tmaskin) GT 0 then begin
   tmask = mdata & tmask[*] = 0.0
   tmask[tmaskin] = 1.0
   tmask = tmask[xvi[0]:xvi[1],yvi[0]:yvi[1]]
   tmask = where(tmask)
endif

; Use a 2nd degree transformation to allow for evolution and foreshortening
; due to solar rotation.

if sqrt(float(n_elements(tmdi2traceguess[*,0]))) MOD 1.0 NE 0.0 then $
   message,'tmdi2traceguess has bad dimension'
pin = tmdi2traceguess[*,0]
qin = tmdi2traceguess[*,1]
n = long(sqrt(n_elements(pin))+0.5)
n9 = (3>n)  ; at least order 2 transformation
p9 = fltarr(n9,n9) & p9[0,0] = reform(pin,n,n) & p9 = reform(p9,n9*n9)
q9 = fltarr(n9,n9) & q9[0,0] = reform(qin,n,n) & q9 = reform(q9,n9*n9)

if 0 then begin
   pout = p9  ; This just uses the initial guess for debugging purposes
   qout = q9
   inew = poly_2d(timage,pout,qout,2,xvr[1]-xvr[0]+1,yvr[1]-yvr[0]+1, $
          cubic=-0.5)
endif else begin
   inew=auto_align_images(timage,rimage,p9,q9,pout,qout,missing=0.0, $
                          tmask=tmask,rmask=rmask,unrestricted=1, $
                          quiet=quiet,_EXTRA=extra)
endelse

; Check the plate scales and rotation assumptions
p4 = (reform(pout,n9,n9))[0:1,0:1]
q4 = (reform(qout,n9,n9))[0:1,0:1]
pq2rss,p4,q4,erot,exscl,eyscl,exshft,eyshft,enrss,/quiet
if abs(exscl*tindex.cdelt1-mdipix_earth) GT 0.02 OR $
   abs(eyscl*tindex.cdelt2-mdipix_earth) GT 0.02 then begin
   message,/info,'Warning: TRACE and MDI plate scales are inconsistent: ' + $
                 string(max([abs(exscl*tindex.cdelt1-mdipix_earth), $
                             abs(eyscl*tindex.cdelt2-mdipix_earth)])) + $
                 ' arcsec'
endif
; A 0.1 degree roll could give an 0.5" error over the TRACE FOV if the MDI
; roll is wrong, so give a warning here.  But, if the TRACE roll is
; the one that is wrong, then the coordinates from this program will
; be correct. However, maps may be wrong if the TRACE roll does not
; get set properly in the map structures so this is still a good place
; to give a warning.  But that is the problem of the mapping software,
; can't fix that here!  
if abs(erot-mindex.solar_p) GT 0.1 then begin  ; Assumes trace roll is 0.0
   message,/info,'Warning: TRACE and MDI roll angles are inconsistent: ' + $
                 string(erot)+' degrees'
endif

mdi2trace=[[pout],[qout]]

; Get Lat/Lon of MDI data

tracedate = tindex.date_obs
tracedate = anytim(tracedate,/yoh)

ddays = int2secarr(anytim2ints(mdidate),anytim2ints(tracedate))/86400.
print,'ddays:',ddays

; get_latlon changes the pixs input to a 2-elements array so protect mdipix
jmdipix = mdipix 

; For MDI  ( (ny-1)/2.0 - y0 ) * im_scale = ycen 

earth_L0 = mindex.earth_L0
obs_L0  = mindex.obs_L0
earth_B0 = mindex.earth_B0
obs_B0  = mindex.obs_B0
earth_R0 = mindex.earth_R0
obs_R0 = mindex.obs_R0

latlon = get_latlon(nx=xvi[1]-xvi[0]+1L, $
                    ny=yvi[1]-yvi[0]+1L, $
                    x0=(xvi[0]+xvi[1])/2.0-mindex.x0, $
                    y0=(yvi[0]+yvi[1])/2.0-mindex.y0, $
                    b0=obs_b0*!dtor, $
                    pang=mindex.solar_p*!dtor, $  ; should be zero
                    pixs=jmdipix, $
                    rad=obs_r0)

; convert center of TRACE restricted FOV to MDI coordinates
 
xtrace = (xvr[1]-xvr[0])/2.0
ytrace = (yvr[1]-yvr[0])/2.0
pq2xy,xtrace,ytrace,mdi2trace[*,0],mdi2trace[*,1],xtrace_mdi,ytrace_mdi

; If differential rotation were zero and the MDI and earth B0,L0 angles were 
; identical, then
; xcen = ((nxtrace-1.)/2.-(xtrace+xvr[0]))*tindex.cdelt1 + (xtrace_mdi+xvi[0]-mindex.x0)*mdipix_earth
; ycen = ((nytrace-1.)/2.-(ytrace+yvr[0]))*tindex.cdelt2 + (ytrace_mdi+yvi[0]-mindex.y0)*mdipix_earth

; What is the lat/cmd of this MDI point?  Remove solar rotation
; here since that piece of the transformation is not a pointing shift.
; Also remove parallax in the longitude.  The parallax in latitude is
; accounted for already since get_latlon has b0 = mindex.obs_b0 but below
; lonlat2xy uses the earth value.

; compute the parallax in longitude: mindex.obs_L0-mindex.earth_L0

value = (obs_L0-earth_L0+180.) MOD 360.
neg = where(value LT 0.,count)
if count GT 0 then value(neg) = value(neg) + 360.
plon = value - 180. ; this is mindex.obs_L0-mindex.earth_L0 in [-180,+180]

lat = bilinear(latlon.latitude,xtrace_mdi,ytrace_mdi)
cmd = bilinear(latlon.cmd,xtrace_mdi,ytrace_mdi)-diff_rot(ddays,lat,/allen) + plon

xy = lonlat2xy([cmd,lat],tracedate) ; x,y with solar rotation removed

new_tindex = tindex
new_tindex.xcen = ((nxtrace-1.)/2.-(xtrace+xvr[0]))*tindex.cdelt1 + xy[0]
new_tindex.ycen = ((nytrace-1.)/2.-(ytrace+yvr[0]))*tindex.cdelt2 + xy[1]
new_tindex.crpix1 = (nxtrace-1.)/2.
new_tindex.crpix2 = (nytrace-1.)/2.
new_tindex.crval1 = new_tindex.xcen
new_tindex.crval2 = new_tindex.ycen

if NOT keyword_set(quiet) then begin
   print,'Old center: ',tindex.xcen,tindex.ycen
   print,'New center: ',new_tindex.xcen,new_tindex.ycen
   print,'TRACE off by ',$
         sqrt((tindex.xcen-new_tindex.xcen)^2 + $
              (tindex.ycen-new_tindex.ycen)^2),' arcsec'
   print
   ;mmap = mk_mdi_map_earth(mindex,mdata)
   ;index2map,new_tindex,tdata,tmap
   ;plot_map,tmap
   ;plot_map,mmap,/over,/drot,level=[1000.,2000.,4000.,6000.,8000.,10000.]
endif

if sqrt((tindex.xcen-new_tindex.xcen)^2 + $
        (tindex.ycen-new_tindex.ycen)^2) GT 20. then $
   message,/info,'WARNING: The pointing change is large.  ' + $
                 'It is possible that the routine converged to an '+$
                 'incorrect solution.  Try re-running and pass the '+$
                 'keywords "/powell,scale=0.1".'

return,new_tindex

end
